=begin
btm_Sketch_Motion.rb

Copyright 2010, btmsketchup@hotmail.com

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Created by •BTM on 14/01/10.

Name:			Sketch Motion
Version:		0.1: 14/01/10
Description:	This File is an animation script for SketchUp.
=end
require 'sketchup.rb'


##############################################################################################


class SketchMotionPlay
#===================================================================================
def activate
#-----CONSTANTS---------------------------------------------------------------------
	@model = Sketchup.active_model
	@view =  @model.active_view
#-----VARIABLES---------------------------------------------------------------------
	@frame_rate = 1.0 / $sm_frames_per_second
#-----PLAY ANIMATION----------------------------------------------------------------
	playAnimation(@view)
end
#===================================================================================
def playAnimation(view)
#-----BACK TO START IF AT END-------------------------------------------------------
	if $sm_current_frame >= $sm_animation_length
		$sm_current_frame = 1
	end
#-----UPDATE EACH FRAME-------------------------------------------------------------
	@frame_timer = UI.start_timer(@frame_rate, true) {
		# Update the frame number: @frame_rate doesn't seem to be working right (always seems to go as fast as it can, as though @frame_rate was 0)				<----- Problem #1
		$sm_current_frame = $sm_current_frame + 1
		
		# Do every animation that takes place here
		SketchMotion.updateFrame
		#if it's the last frame, stop playing, or loop to 1st frame
		if $sm_current_frame >= $sm_animation_length
			if $sm_animation_loop == false
				UI.stop_timer @frame_timer
			elsif $sm_animation_loop == true
				$sm_current_frame = 1
			end
		end
	}
end
#===================================================================================
def deactivate(view)
	#if the animation's running, stop it
	if @frame_timer
		UI.stop_timer @frame_timer
	end
end
#===================================================================================
end
UI.menu("plugins").add_item ("Sketch Motion Play") { Sketchup.active_model.select_tool SketchMotionPlay.new}


##############################################################################################


class SketchMotionExport
#===================================================================================
def activate
	prompts = ["Destination", "Title"]
	defaults = ["", "animation"]
	list =  ["", ""]
	results = UI.inputbox prompts, defaults, list, "Destination folder for images"
	
	if results[0] != nil
		for n in 1...$sm_animation_length do
			$sm_current_frame = n
			SketchMotion.updateFrame
			name = results[0] + "/" + results[1] + " " + n.to_s + ".png"
			puts name.to_s
			image = Sketchup.active_model.active_view.write_image name
		end
	end
end
#===================================================================================
end
UI.menu("plugins").add_item ("Sketch Motion Export") { Sketchup.active_model.select_tool SketchMotionExport.new}


##############################################################################################


class SketchMotionCurrentSettings
#=====Animation Settings============================================================
$sm_frames_per_second = 25.0
$sm_animation_length = 100
$sm_current_frame = 1
$sm_animation_loop = false
#=====Animated Entities=============================================================
$sm_animated_groups = []
$sm_animated_vertices = []
#===================================================================================
def activate
	if $sm_animation_loop == true
		looped = "yes"
	else
		looped = "no"
	end
	
	prompts = ["FPS", "Length(frames)", "Cur. Frame", "Loop?"]
	defaults = [$sm_frames_per_second, $sm_animation_length, $sm_current_frame, looped]
	list =  ["", "", "", "yes|no" ]
	results = UI.inputbox prompts, defaults, list, "Please input desired Settings."
	
	$sm_frames_per_second = results[0]
	$sm_animation_length = results[1]
	$sm_current_frame = results[2]
	
	if results[3] == "yes"
		$sm_animation_loop = true
	else
		$sm_animation_loop = false
	end
		
	#save the model's framerate + animation length + whether or not to loop, to get when we open it
	Sketchup.active_model.set_attribute "sketch motion animation", "length", $sm_animation_length
	Sketchup.active_model.set_attribute "sketch motion animation", "fps", $sm_frames_per_second
	Sketchup.active_model.set_attribute "sketch motion animation", "loop", $sm_animation_loop
	
	SketchMotion.updateFrame
end
#===================================================================================
end
UI.menu("plugins").add_item ("Sketch Motion Settings") { Sketchup.active_model.select_tool SketchMotionCurrentSettings.new}


###### OBSERVER ##############################################################################


class SketchMotionLoadCollections < Sketchup::AppObserver
#===================================================================================
def onOpenModel(model)
	#get the length + framerate of the animation from the last time it was opened
	#it would be possible to save multiple animations by changing each one's attribute dictionary… I should do that… Later…
	if Sketchup.active_model.get_attribute "sketch motion animation", "length"
		$sm_animation_length = Sketchup.active_model.get_attribute "sketch motion animation", "length"
	end
	if Sketchup.active_model.get_attribute "sketch motion animation", "fps"
		$sm_frames_per_second = Sketchup.active_model.get_attribute "sketch motion animation", "fps"
	end
	if Sketchup.active_model.get_attribute "sketch motion animation", "loop"
		$sm_animation_loop = Sketchup.active_model.get_attribute "sketch motion animation", "loop"
	end
	
	vertices = []
	#find all the animated entities, and put them back to the arrays they were in
	model.entities.each { |e|
		if e.is_a? Sketchup::ComponentInstance or e.is_a? Sketchup::Group
			if e.attribute_dictionary "sketch motion key"
				$sm_animated_groups << e
			end
		end
		if e.is_a? Sketchup::Edge
			vertices << e.start
			vertices << e.end
		end
	}
	vertices.uniq!
	
	vertices.each { |v|
		if v.attribute_dictionary "sketch motion key"
			$sm_animated_vertices << v
		end
	}
end
#===================================================================================
def onNewNodel(model)
	#reset the animations for the new model
	$sm_animated_groups = []
	$sm_animated_vertices = []
end
#===================================================================================
end
Sketchup.add_observer(SketchMotionLoadCollections.new)


##############################################################################################


class SketchMotionSetKey
#===================================================================================
def activate
	selection = Sketchup.active_model.selection
	
	selection.each { |s|
		#if it's a group, save it's transformation and add it to the proper array, if it's not already there
		if s.is_a? Sketchup::ComponentInstance or s.is_a? Sketchup::Group
			s.set_attribute "sketch motion key", $sm_current_frame.to_s, [$sm_current_frame, s.transformation.to_a]
			if not $sm_animated_groups.index(s)
				$sm_animated_groups << s
			end
		end
		#if it's an edge / face, save the vertices positions and add it to the proper array, if it's not already there
		if s.is_a? Sketchup::Edge or s.is_a? Sketchup::Face
			s.vertices.each { |vertex|
				vertex.set_attribute "sketch motion key", $sm_current_frame.to_s, [$sm_current_frame, vertex.position.to_a]
				if not $sm_animated_vertices.index(vertex)
					$sm_animated_vertices << vertex
				end
			}
		end
	}
end
#===================================================================================
end
UI.menu("plugins").add_item ("Sketch Motion Set Key") { Sketchup.active_model.select_tool SketchMotionSetKey.new}


##############################################################################################


class SketchMotion
#===================================================================================
def SketchMotion.updateFrame
	model = Sketchup.active_model
	view = model.active_view
	#animate groups. there's no camera animation yet, linear interpolation sucks
	$sm_animated_groups.each { |group|
	if group.valid?
		#find the previous and next frames
		keys = SketchMotion.prevNextKeys (group)
		if keys[0] and keys[1]
			#if there was a previous key and another key yet to come, smooth into that key's transformation
			#I should find a better way to smooth than linear interpolation, since it doesn't always look good						<----- Problem #2
			interpolation_value = ($sm_current_frame.to_f - keys[0][0].to_f) / (keys[1][0].to_f - keys[0][0].to_f)
			t1 = Geom::Transformation.new keys[0][1]
			t2 = Geom::Transformation.new keys[1][1]
			new_transformation = Geom::Transformation.interpolate t1, t2, interpolation_value
			group.transformation = new_transformation
		elsif keys[0] and (keys[1] == false)
			#otherwise, just use the current / previous frame's transformation
			group.transformation = keys[0][1]
		elsif keys[1] and (keys[0] == false)
			#otherwise, just use the next frame's transformation
			group.transformation = keys[1][1]
		end
	end
	}
	#animate vertices. really glitchy if used inside of groups
	$sm_animated_vertices.each { |vertex|
	if vertex.valid?
		keys = SketchMotion.prevNextKeys (vertex)
		if keys[0] and keys[1]
			#if there was a previous key and another key yet to come, smooth into that key's position
			#I should find a better way to smooth than linear combination, since it doesn't always look good						<----- Problem #3
			interpolation_value = ($sm_current_frame.to_f - keys[0][0].to_f) / (keys[1][0].to_f - keys[0][0].to_f)
			interpolation_value_2 = 1 - interpolation_value
			p1 = Geom::Vector3d.new keys[0][1]
			p2 = Geom::Vector3d.new keys[1][1]
			position_vector = Geom::Vector3d.linear_combination interpolation_value_2, p1, interpolation_value, p2
			vector = position_vector  - Geom::Vector3d.new (vertex.position.to_a)
			Sketchup.active_model.entities.transform_entities vector, vertex
		elsif keys[0] and (keys[1] == false)
			#otherwise, just use the current / previous frame's position
			vector = Geom::Vector3d.new keys[0][1] - Geom::Vector3d.new (vertex.position.to_a)
			Sketchup.active_model.entities.transform_entities vector, vertex
		elsif keys[1] and (keys[0] == false)
			#otherwise, just use the next frame's position
			vector = Geom::Vector3d.new keys[1][1] - Geom::Vector3d.new (vertex.position.to_a)
			Sketchup.active_model.entities.transform_entities vector, vertex
		end
	end
	}
	#reset the view
	view.invalidate
end
#=================================================================================== 

def SketchMotion.prevNextKeys(entity)
	#hopefully no one tries to make a 1000000000 frame long animation
	previous_frame = 0
	next_frame = 999999999
	previous_key = nil
	next_key = nil
	entity_keys = entity.attribute_dictionary "sketch motion key"
	#go through the keys to find the closest one below / on the current frame, and the next one
	entity_keys.values.each { |k|
		frame = k[0]
		if frame > previous_frame and frame <= $sm_current_frame
			previous_frame = frame
			previous_key = k
		end
		if frame < next_frame and frame >  $sm_current_frame
			next_frame = frame
			next_key = k
		end
	}
	return [previous_key, next_key]
end
#===================================================================================
end
